/*
 * linux/fs/ext2/ioctl.c
 *
 * Copyright (C) 1993, 1994, 1995
 * Remy Card (card@masi.ibp.fr)
 * Laboratoire MASI - Institut Blaise Pascal
 * Universite Pierre et Marie Curie (Paris VI)
 */

#include <linux/fs.h>
#include <linux/ext2_fs.h>
#include <linux/sched.h>
#include <asm/uaccess.h>

/* RSBAC */
#ifdef CONFIG_RSBAC
#include <rsbac/adf.h>
#endif

int ext2_ioctl (struct inode * inode, struct file * filp, unsigned int cmd,
		unsigned long arg)
{
	unsigned int flags;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_adf_request_t     rsbac_request;
        enum  rsbac_target_t          rsbac_target = T_NONE;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        rsbac_pr_debug(aef, "calling ADF\n");
        switch (cmd)
          {
            case EXT2_IOC_GETFLAGS:
            case EXT2_IOC_GETVERSION:
              rsbac_request = R_GET_PERMISSIONS_DATA;
              break;
            case EXT2_IOC_SETFLAGS:
            case EXT2_IOC_SETVERSION:
              rsbac_request = R_MODIFY_PERMISSIONS_DATA;
              break;
            default:
              rsbac_request = R_NONE;
	  }
        if(S_ISSOCK(inode->i_mode))
          {
            #ifdef CONFIG_RSBAC_NET_OBJ
            #if !defined(CONFIG_RSBAC_NET_OBJ_UNIX)
            if(   inode->u.socket_i.ops
               && (inode->u.socket_i.ops->family != AF_UNIX)
              )
            #endif
              {
                rsbac_target = T_NETOBJ;
                rsbac_target_id.netobj.sock_p
                  = &inode->u.socket_i;
                rsbac_target_id.netobj.local_addr = NULL;
                rsbac_target_id.netobj.local_len = 0;
                rsbac_target_id.netobj.remote_addr = NULL;
                rsbac_target_id.netobj.remote_len = 0;
              }
            #endif
          }
        else
          {
            if (S_ISDIR(inode->i_mode))
              rsbac_target = T_DIR;
            else if (S_ISFIFO(inode->i_mode))
              rsbac_target = T_FIFO;
            else if (S_ISLNK(inode->i_mode))
              rsbac_target = T_SYMLINK;
            else
              rsbac_target = T_FILE;
            rsbac_target_id.file.device = inode->i_dev;
            rsbac_target_id.file.inode  = inode->i_ino;
            rsbac_target_id.file.dentry_p = filp->f_dentry;
	  }
        rsbac_attribute_value.ioctl_cmd = cmd;
        if(   (rsbac_request != R_NONE)
           && !rsbac_adf_request(rsbac_request,
                                 current->pid,
                                 rsbac_target,
                                 rsbac_target_id,
                                 A_ioctl_cmd,
                                 rsbac_attribute_value))
          {
            return -EPERM;
          }
        #endif

	ext2_debug ("cmd = %u, arg = %lu\n", cmd, arg);

	switch (cmd) {
	case EXT2_IOC_GETFLAGS:
		flags = inode->u.ext2_i.i_flags & EXT2_FL_USER_VISIBLE;
		return put_user(flags, (int *) arg);
	case EXT2_IOC_SETFLAGS: {
		unsigned int oldflags;

		if (IS_RDONLY(inode))
			return -EROFS;

		if ((current->fsuid != inode->i_uid) && !capable(CAP_FOWNER))
			return -EACCES;

		if (get_user(flags, (int *) arg))
			return -EFAULT;

		oldflags = inode->u.ext2_i.i_flags;

		/*
		 * The IMMUTABLE and APPEND_ONLY flags can only be changed by
		 * the relevant capability.
		 *
		 * This test looks nicer. Thanks to Pauline Middelink
		 */
		if ((flags ^ oldflags) & (EXT2_APPEND_FL | EXT2_IMMUTABLE_FL)) {
			if (!capable(CAP_LINUX_IMMUTABLE))
				return -EPERM;
		}

		flags = flags & EXT2_FL_USER_MODIFIABLE;
		flags |= oldflags & ~EXT2_FL_USER_MODIFIABLE;
		inode->u.ext2_i.i_flags = flags;

		ext2_set_inode_flags(inode);
		inode->i_ctime = CURRENT_TIME;
		mark_inode_dirty(inode);
		return 0;
	}
	case EXT2_IOC_GETVERSION:
		return put_user(inode->i_generation, (int *) arg);
	case EXT2_IOC_SETVERSION:
		if ((current->fsuid != inode->i_uid) && !capable(CAP_FOWNER))
			return -EPERM;
		if (IS_RDONLY(inode))
			return -EROFS;
		if (get_user(inode->i_generation, (int *) arg))
			return -EFAULT;	
		inode->i_ctime = CURRENT_TIME;
		mark_inode_dirty(inode);
		return 0;
	default:
		return -ENOTTY;
	}
}
